{% extends "base.html" %}

{% block title %}
	&raquo; MDD Converter
{% endblock %}

{% block nav %}
	&raquo; MDD Converter
{% endblock %}

{% block content %}
	<!-- Layer Section -->
	<div class="layer-section">
		<!-- Section Header -->
		<div class="section-header">
			<h2>MDD Converter</h2>	
		</div>
		<!-- End Section Header -->
		<!-- Section Content -->
		<div class="section-content">
			<div class="section-data">
				<p>
					The <a
					href="https://www.elexon.co.uk/reference/technical-operations/market-domain-data/"
							>Market
					Domain Data</a> (MDD) changes over time, and so needs to be regularly
					updated. Unfortunately Elexon don't provide the MDD for automated
					download, so you have to manually download it now and again and
					manually import it into Chellow. The MDD has a version number, so you
					only need to do an update if the version number has been incremented.
					The form below will take a MDD ZIP file and convert it into the
					General Import format so that it can be
					<a href="/general_imports">imported</a> into Chellow. At the moment,
					the files within the zip that are recognized are:
				</p>
				<ul>
					<li>Market_Participant</li>
					<li>Market_Role</li>
					<li>Market_Participant_Role</li>
					<li>Line_Loss_Factor_Class</li>
					<li>MTC_Meter_Type</li>
					<li>MTC_in_PES_Area</li>
					<li>Meter_Timeswitch_Class</li>
				</ul>
			<div class="standardForm">
				<form
						enctype="multipart/form-data" action="/reports/163" method="post">
					<fieldset>
						<legend>Convert an MDD CSV file</legend>
							<input type="file" name="file">
							<input type="submit" value="Convert">
					</fieldset>
				</form>
			</div>
		</div>
		<!-- End Section Content -->
	</div>
	<!-- End Layer Section -->
{% endblock %}
